# Copyright (C) 2001 by Dr. Dieter Maurer <dieter@handshake.de>
# D-66386 St. Ingbert, Eichendorffstr. 23, Germany
#
#			All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice and this permission
# notice appear in all copies, modified copies and in
# supporting documentation.
# 
# Dieter Maurer DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL Dieter Maurer
# BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
# DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
# PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.
'''Form Dispatcher

 A form dispatcher is used to manage the variables of a form.
 It provides services for form initialization, form update
 and the dispatch out of the form to various form actions.
 During dispatch, the current form values are saved in a session
 object. When the form is updated, the data is put back
 into 'REQUEST.form' and updated with the new values provided
 by 'update'.
 
 A form dispatcher works with (at least) the following objects:
 
   Client -- the id of the form controlled by the dispatcher
 
       The dispatcher calls this object like a
       DTML method (or through 'mapply') after 'REQUEST.form'
       and 'REQUEST.other' have been set up.
 
       The default is 'Client'.
 
   ClientDescription -- the description of the managed form
       fields.
 
       The dispatcher calls this object like a DTML method.
       The call should return either a list of form value descriptions
       or a dictionary thereof. Each description is a dictionary
       describing various aspects managed by the dispatcher.
 
       The default is 'ClientDescription'
 
    SessionObject -- the id of a session managing object.
 
       This object is used to save the current form values during
       dispatch.
 
       The default is 'FSSession'.
 
    SessionKey -- the key to use for saving the current form values
       in the session object.
 
       The default is the dispatchers physical path.
 
 Use Cases
 
   The form dispatcher supports the following use cases:
 
     'new' -- the form is used to create a new object.
 
         The initial values of form variables are not taken from
         the environment unless the form description contains
         an explicit 'init' or 'fixed' description to this effect.
         Other form variables are empty (depending on 'dtype').
 
     'edit' -- the form is used to edit an existing object.
 
         Most form variable values are taken from the environment.
         Only 'fixed' values are taken from the description.
 
     'show' -- render with currently remembered values.

         This method is used, when the Dispatcher is part
         or a larger context, e.g. a Wizzard.
         This context has performed global initialization or
         updates and then selectively calls for rendering
         of a dispatcher.

     'clear' -- the form values are reset to empty (depending on 'dtype').
         Only 'fixed' values are not changed.
 
     dispatch -- the form is left with the intention to come back
         with new or changed values. The current form values are
         saved in the session object to be restored lated when
         we come back.
 
     'update' -- update the form after a dispatch.
 
         The original form values are restored from the session object.
         They can be overridden or completed by additional values provided in
         the 'update' call.
 
        
 Event Sequences
 
   The only anticipated event sequences are::
 
     (new | edit) (clear | show | dispatch update)* *leave* ?
 
 
 Properties
 
   'Client__' -- the id of the form controlled by the dispatcher
 
     The default is 'Client'.
 
   'ClientDescription__' -- the id of a method returning a dictionary or list that describes the form variables.
 
     The default is 'ClientDescription'.
 
     Each variable is described by a dictionary.
     The following keys are interpreted. There may be more, used for
     other purposes:
 
       'name' -- the variable name
 
       'init' -- an initial value for 'new' calls.
 
          This is either 'None' or a value.
          'None' means, use from the environment (if present).
          A missing 'init' is equivalent to the type
          specific empty value.
 
       'fixed' -- the value is not changed by 'clear' or '__call__'
 
          This is either 'None' or a value.
          'None' means, use from the environment (if present).
 
          If 'fixed' is not present, then 'clear' resets the variable
          to the type specific empty value.

       'permanent' -- the value is not changed by 'update'.

          This should be a truth value, the default is '0'.

          It is usually used together with 'fixed' variables
          to prevent modification by 'update'.

          It currently only affects 'update'.
          It might be more intuitive, if 'permanent' would imply 'fixed'.
 
 
       'dtype' -- the variables data type (used to determine empty values)
 
          Must be either 'string', 'int', 'list' or 'tuple'.
          Defaults to 'string'.
 
 
   'SessionObject__' -- the id of the session object used to remember the values
 
     The default is 'FSSession'.
     
     It is expected, that the session object has 'get' and 'set'
     methods to retrieve and set session data.
 
 
   'SessionKey__' -- the saved values are collected in a dictionary and saved under this key.
 
     The default is the physical path of the dispatcher.

     If the value is '*dynamic*', then the session key is
     dynamically computed form the physical path.
     This is necessary, if the dispatcher is used as part
     of a ZClass.
 
 
   'initializeSession__' -- the id of a method that initialized the session.
 
      The default is 'initializeSession'.
 
      'FSSession' requires that the session is initialized precisely
      once in each request that uses the session.
      Initializing more than once may lead to loss of data.
      Not initializing may lead to exceptions, let the application
      work with stale data and does not save modifications.
      It is expected that 'initializeSession' is idempotent,
      i.e. can be called repeatedly in a single request without
      advertant effects.
 
      For some session types (unlike FSSession), no explicit session
      initialization may be required. In such a case,
      'initializeSession__' may be deleted.
 
 
 Methods
 
   'new'
 
       fill a new form (mostly empty).
 
       All variables get their initialization/fixed values (if defined).
       Other variables get the 'dtype' specific empty value.
 
       The values are placed in 'REQUEST' and 'REQUEST.form'.
 
       'None' values for 'init' or 'fixed' are resolved from
       the environment.
 
       The special value 'edit__=0' is defined.
 
       The method is called either directly from the Web
       or from DTML, to better control the environment.
 
 
     Arguments: DTML-like
 
 
   'edit'
 
     fill a form for editing (mostly filled).
 
     All variables get their values from the environment
     with the exception of 'fixed' variables.
     They get the specified value, unless the value
     is 'None'.
 
     The values are placed in 'REQUEST' and 'REQUEST.form'.
 
     The special value 'edit__=1' is defined.
 
     The method is usually called
     from DTML which sets up the environment.
 
     Arguments: DTML-like
 
 
   'update'
 
     update the remembered values and render the client.
 
     The variables are filled with their remembered values.
     New values from 'update' override pervious values,
     unless 'errors__' has a true value.
     In this latter case, no values are overridden.
     In no case, 'update' overrides a variable declared
     as 'permanent'.
 
     Usually, the update values are taken from the environment.
     If, however, 'values__' is defined, then its value
     must be a mapping and it, alone, is used for updating.
     The namespace is not relevant in this case.
 
     The values are placed in 'REQUEST' and 'REQUEST.form'.
 
     'update' restores the base URL, remembered during the
     last dispatch. This is in order to counter the lengthening
     URL problem that is favoured by Zopes acquisition.
 
     The method is usually called from DTML. It may, however,
     also be called directly from the Web.
 
 
     Arguments
 
       'update' accepts DTML-like arguments.
       However, 'errors__' and 'values__' are treated
       specially.
 
       If 'errors__' is defined and has a true value, the
       old values are simply restored. No values are
       overridden by the 'update'.
 
       If 'values__' is defined, then (with the exception
       of 'errors__'), the environment is irrelevant.
       Only 'values__' (which must be a mapping) defines
       values to override remembered values.
       This provides for precise control over the overridden
       values.
 
 
   '__call__'
 
     remember the current values in the session object (with the exception
     of 'fixed' variables)
     and dispatch to a new destination.
 
     If the environment does not contain a value for a
     variable which is not 'fixed',
     then the empty value is used. This
     is necessary for checkboxes and multiple selections to work
     correctly.
 
     The values and the base URL are saved in the session object
     for later restoration by 'update'.
 
     To determine the destination, '__call__' looks for
     a variable of the form '(.*)__disp(.[xy])?'.
     It takes '\\1' as the destination.
 
     If the destination contains '?', it is split at it.
     The first part becomes the destination, the second part
     is decoded as a query string and the corresponding
     parameters placed into 'REQUEST' and 'REQUEST.form'.

     The destination may be an URL path. 'restrictedTraverse'
     is used to find the actual destination.

     If no destination can be found, the client is simply
     rendered (equivalent to 'show'). This is useful
     to have shorter URLs. Currently, the same happens,
     when 'REQUEST' does not have a 'form' attribute.
     However, this may change in the future.
 
     The method is usually called only from the Web.
 
     Arguments: dtml-like
 
 
   'clear'
 
     clear (make empty) non-fixed fields and render the client.
 
     Arguments: DTML-like


   'show'

     render the client.

     Arguments: DTML-like
 
 
   'getValues'
 
     return a mapping object containing the saved values.
 
     This is the mapping saved in the session object.
     Writing to it changes the session maintained information.
 
     Arguments: none
 
 
 Arguments
 
   Most methods support DTML-like arguments.
   This means, there are two optional positional arguments
   'client' and 'REQUEST' and arbitrary keyword arguments.
   Unlike DTML-objects, the default value for 'client' is
   the dispatcher object. 'REQUEST' may be a mapping
   or a DTML namespace.
 
   The methods construct a DTML namespace from this in the same
   way, DTML objects do. When an object is rendered,
   the 'render' method of this namespace is used.


 Rendering

   Usually, 'new', 'edit', 'clear', and 'show' render
   the client. If, however, 'DispatcherDoNotRender__'
   is true, rendering and 'REQUEST' modification
   are suppressed. This allows e.g.
   a Wizzard to call these methods just to change the remembered values.

   For a DTML object or when the object is not acquisition
   wrapped or when the namespace is bound,
   the namespace''s 'render' method is used
   for rendering. Other objects are rendered
   through 'ZPublisher.mapply.mapply'.

   Zopes acquisition and the possibility to call methods both
   via HTTP and directly can lead to great confusion with
   respect to base URLs. The base URL is essential for
   the resolution of any relative URL in an HTML page.
   The Dispatcher tries to counter this confusion by
   explicitly setting the HTML base element.
   For this purpose, it must find out its own URL.
   Note, that it can not use 'absolute_url', as this method
   discards any context.
   For 'new', 'edit', 'clear', 'show' and '__call__', the dispatcher
   assumes that it was activated directly by HTTP and that it
   can use 'REQUEST.URL1' or 'REQUEST.URL0', respectively.
   If this assumption is wrong,
   the caller can define 'DispatcherURL__'. It should be
   the URL of the dispatcher object.
   If the client is rendered at the end of an
   'update', the dispatcher restores the remembered base URL.

   **Note** that the URL/BASE request variables
   are fixed for the rendering as well. They reflect the
   true client/destination information, not that of the dispatcher.
   Other request variables are not changed. This may change in the future.
   '''



from OFS.Folder import Folder
from AccessControl.SecurityInfo import ClassSecurityInfo
from DocumentTemplate.DT_String import TemplateDict, InstanceDict
from ZPublisher.HTTPRequest import HTTPRequest
from Globals import InitializeClass, DTMLFile

from string import join, split, rfind
from re import compile

from HelpSupport import HelpSTXTopic_S

def getDispatcherHelp():
  DispatcherHelp= HelpSTXTopic_S('DispatcherHelp', 'Dispatcher Documentation',__doc__)
  return DispatcherHelp


_matchDest= compile(r'(.+)__disp(?:\.[xy])?$').match

class Dispatcher(Folder):
  '''A dispatcher manages the variables of a form.
  For details, see the <a href="/Control_Panel/Products/Dispatcher/Help/DispatcherHelp">help</a>.'''
  meta_type= 'Form Dispatcher'

  _properties= Folder._properties + (
    {'id': 'Client__', 'type': 'string', 'mode':'w',},
    {'id': 'ClientDescription__', 'type': 'string', 'mode':'w',},
    {'id': 'SessionObject__', 'type': 'string', 'mode':'w',},
    {'id': 'SessionKey__', 'type': 'string', 'mode':'w',},
    {'id': 'initializeSession__', 'type': 'string', 'mode':'wd',},
    )
  Client__= 'Client'
  ClientDescription__= 'ClientDescription'
  SessionObject__= 'FSSession'
  SessionKey__= '' # equivalent to the objects physical path
  initializeSession__= 'initializeSession'

  manage_options= (
    ({'label' : 'Contents', 'action':'manage_main',
      'help' : ('Dispatcher','DispatcherHelp')},)
    + Folder.manage_options[1:]
    )

  _security= ClassSecurityInfo()

  _security.declareProtected('View', '__call__',
                             'new', 'edit', 'clear', 'update',
                             'getValues', 'value')


  def new(self, client=None, REQUEST=None, **kw):
    '''fill new form (most variables empty).'''
    (ns,level,desc,REQUEST,fields)= self._setup(client,REQUEST,kw)
    marker= fields= {'edit__' : 0,}
    for d in desc:
      v= d.get('fixed', d.get('init',marker))
      if v is marker: v= self._selectEmpty(d)
      elif v is None: v= self._selectValue(d,ns)
      fields[d['name']]= v
    return self._finish(ns, level, fields, REQUEST)
  newisDocTemp= 1 # behave like a Document Template


  def edit(self, client=None, REQUEST=None, **kw):
    '''edit existing object (most variables with value).'''
    (ns,level,desc,REQUEST,fields)= self._setup(client,REQUEST,kw)
    fields= {'edit__' : 1,}
    for d in desc:
      fields[d['name']]= self._selectValue(d,ns)
    return self._finish(ns, level, fields, REQUEST)
  editisDocTemp= 1 # behave like a Document Template


  def clear(self, client=None, REQUEST=None, **kw):
    '''clear non-fixed variables; retain fixed values.'''
    (ns,level,desc,REQUEST,fields)= self._setup(client,REQUEST,kw)
    marker= fields
    for d in desc:
      n= d['name']
      if not d.has_key('fixed'): fields[n]= self._selectEmpty(d)
    return self._finish(ns, level, fields, REQUEST)
  clearisDocTemp= 1 # behave like a Document Template


  def __call__(self, client=None, REQUEST=None, **kw):
    '''dispatch.'''
    #import pdb; pdb.set_trace()
    if not hasattr(REQUEST,'form'):
      # not called via Web - render
      # Note: we may change this to render only, if
      #       destination is missing
      return self._render(client,REQUEST,0,kw)
    dest= None; form= REQUEST.form; match= _matchDest
    for v in form.keys():
      m= match(v)
      if m is not None: dest= m.group(1); del form[v]; break
    if dest is None:
      dest= form.get('__disp')
    if form.has_key('__disp'): del form['__disp']
    if not dest: return self._render(client,REQUEST,0,kw)
    args= None
    if '?' in dest: 
      (dest,qs)= split(dest,'?',1)
      args= self._decodeQueryString(qs)
    return self._saveAndRender(REQUEST, dest, args)
  # DM: the following is ineffective
  #     we would need to place the "isDocTemp" on the
  #     object, but we do not want to "render" dispatchers
  #     as a Document Template, as "mapply" is much safer
  #__call__isDocTemp= 1 # behave like a Document Template


  def _saveAndRender(self, REQUEST, client=None, args=None, **kw):
    '''save the values from 'REQUEST' and render 'dest'.'''
    (ns,level,desc,REQUEST,fields)= self._setup(None,REQUEST,kw)
    marker= fields
    for d in desc:
      n= d['name']
      if not d.has_key('fixed'):
        v= REQUEST.form.get(n, marker)
        if v is marker: v= self._selectEmpty(d)
        fields[n]= v
    return self._finish(ns, level, fields, REQUEST,
                        client=client, args=args, urlLevel= 0)


  def update(self, client=None, REQUEST=None, **kw):
    '''update with fresh values.'''
    (ns,level,desc,REQUEST,fields)= self._setup(client,REQUEST,kw)
    marker= fields; values= ns
    has= ns.has_key
    if not (has('errors__') and ns['errors__']):
      if has('values__'): values= ns['values__']; has= values.has_key
      for d in desc:
        n= d['name']
        if has(n) and not d.get('permanent'): fields[n]= values[n]
    return self._finish(ns, level, fields, REQUEST, urlLevel=None)
  updateisDocTemp= 1 # behave like a Document Template


  def show(self, client=None, REQUEST=None, **kw):
    '''render the client.'''
    return self._render(client, REQUEST, 1, kw)
  showisDocTemp= 1 # behave like a Document Template



  def getValues(self):
    '''mapping object containing the remembered values.'''
    S= getattr(self,self.SessionObject__)
    iS= getattr(self.aq_base,'initializeSession__',None)
    if iS is not None:
      if not iS: iS= 'initializeSession'
      iS= getattr(self,iS)
      iS(S)
    sK= self._getSessionKey()
    f= None
    try: f= S[sK]
    except KeyError: pass
    if f is None: f= {}; S.set(sK,f)
    return f

  def value(self, client=None, REQUEST=None, **kw):
    '''Do not use!
    Implemented for backward compatibility with an old version.

    Updates the saved values with *values__*, if present.
    Puts the value of *varname__* in 'REQUEST' if given,
    otherwise the complete session object under the name 'fields'.
    '''
    (ns,level,desc,REQUEST,fields)= self._setup(client,REQUEST,kw)
    values= ns.has_key('values__') and ns['values__']
    if values: fields.update(values)
    varname= ns.has_key('varname__') and ns['varname__']
    if varname:
      REQUEST.set(varname, fields.get(varname))
    else:
      REQUEST.set('fields', fields)
    if level: ns._pop(level)
    


  def _setup(self,client,REQUEST,kw):
    if client is None: client= self
    if REQUEST is None: REQUEST= self.REQUEST
    pushed= None; ns= REQUEST
    try:
      if ns.__class__ is TemplateDict: pushed= 0
    except AttributeError: pass
    if pushed is None:
      ns= TemplateDict(); ns.validate= None; pushed= 0
      if REQUEST: ns._push(REQUEST); pushed= 1
    if client: ns._push(InstanceDict(client,ns)); pushed= pushed+1
    if kw: ns._push(kw); pushed= pushed+1
    return ns, pushed, self._getDescription(ns), self.REQUEST, self.getValues()


  def _finish(self,ns,level,fields,REQUEST,
              client=None, args=None, urlLevel=1):
    getattr(self,self.SessionObject__).set(self._getSessionKey(),fields)
    try:
      if ns.has_key('DispatcherDoNotRender__') and ns['DispatcherDoNotRender__']:
        return
      if urlLevel is not None:
        self._rememberBaseURL(ns,fields,REQUEST,urlLevel)
      if client is None: client= self.Client__
      clientURL= fields['ClientURL__']= '%s/%s' % (fields['BaseURL__'],client)
      REQUEST.form.update(fields); REQUEST.other.update(fields)
      if args: REQUEST.form.update(args); REQUEST.other.update(args)
      self._resetURLInfo(clientURL,REQUEST)
      REQUEST.RESPONSE.setBase(REQUEST.URL1)
      C= self.restrictedTraverse(client); B= getattr(C,'aq_base',C)
      if getattr(B,'isDocTemp',0) or B is C \
         or (getattr(C,'__render_with_namespace__',0) and 
             C.getBindingAssignments().isNameAssigned('name_ns')):
        return ns.render(C)
      else: return magicApply(C,REQUEST)
    finally:
      if level: ns._pop(level)


  def _render(self, client, REQUEST, urlLevel, kw):
    (ns,level,desc,REQUEST,fields)= self._setup(client,REQUEST,kw)
    return self._finish(ns, level, fields, REQUEST, urlLevel= urlLevel)


  def _getDescription(self,ns):
    '''return a list of descriptions.'''
    d= ns.render(getattr(self,self.ClientDescription__))
    items= getattr(d,'items',None)
    if items:
      l= []
      for (k,v) in items():
        v['name']= k
        l.append(v)
      return l
    return d


  def _selectEmpty(self, d,
                   map= { 'string' : '',
                          'int' : 0,
                          'list' : [],
                          'tuple' : (),
                          }):
    t= d.get('dtype','string')
    return map[t]


  def _selectValue(self,d,ns):
    name= d['name']
    if ns.has_key(name): return ns[name]
    return self._selectEmpty(d)


  def _decodeQueryString(self,qs):
    r= HTTPRequest(None,
                   {'QUERY_STRING': qs,
                    'SERVER_URL': '',
                    },
                   None, 1)
    r.processInputs()
    return r.form

  def _rememberBaseURL(self,ns,fields,REQUEST,urlLevel):
    if ns.has_key('DispatcherURL__'):
      base= ns['DispatcherURL__']
    else:
      base= REQUEST['URL%d' % urlLevel]
    fields['BaseURL__']= base

  def _resetURLInfo(self,url,REQUEST):
    '''reset *URL* and *BASE* variables of *REQUEST* to match *url*.

    **ATTENTION** This is hacky. May work only for Zope 2.3,
    as it uses undocumented internal details about *URL*
    and *BASE* computation.'''
    server= join([REQUEST.SERVER_URL] + REQUEST._script + [''],'/')
    sn= len(server)
    if url[:sn] != server:
      raise ValueError, 'Dispatcher._resetURLInfo: URL %s does not match server URL %s' % (url,server)
    url= url[sn:]
    steps= []
    for s in split(url,'/'):
      if s == '.': pass
      elif s == '..': steps.pop()
      else: steps.append(s)
    REQUEST._steps= steps
    REQUEST._resetURLS()


  def _getSessionKey(self):
    sK= self.SessionKey__
    if not sK: sK= self.SessionKey__= join(self.getPhysicalPath(),'/')
    if sK == '*dynamic*':
      return join(self.getPhysicalPath(),'/')
    return sK



  # do not let ZPublisher acquire "index_html"
  index_html= None


InitializeClass(Dispatcher)
    

def manage_addDispatcher(self,id,REQUEST=None):
  '''constructor'''
  self= self.aq_parent.aq_parent # skip Dispatcher(Product) and ProductDispatcher
  o= Dispatcher()
  o.id= id
  self._setObject(id,o)
  if REQUEST is not None:
    o= self._getOb(id)
    REQUEST.RESPONSE.redirect(o.absolute_url() + '/manage_propertiesForm')


manage_addDispatcherForm= DTMLFile('www/DispatcherAdd', globals())


# mapply infrastructure
from ZPublisher.Publish import mapply, dont_publish_class

def _missing_name(name, context):
  if name == 'self': return context.aq_parent
  context.REQUEST.response.badRequestError(name)

def magicApply(object,request,
               mapply= mapply,
               missing_name= _missing_name,
               ):
  return mapply(object, (), request,
                maybe=1,
                missing_name= missing_name,
                handle_class= dont_publish_class,
                context= object, bind=1)
